/* Seamless3d Server

Copyright (C) 2002-2009 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

//g++ seamless3d_server.cpp var.cpp -o seamless3d_server
//./seamless3d_server http://www.seamless3d.com

#define MAX_CLIENT 105

#define FD_SETSIZE MAX_CLIENT

float g_version=2.157f;

#define PI                  3.14159265358979323846

#define PORT 7778

//#define POSIX 1

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
#include <winsock2.h>
#else
#define POSIX 1
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#endif

int g_numOfClients = 0;

#pragma warning( disable : 4996 4305)

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "var.h"


#ifdef WIN32
inline void close(SOCKET& socket){
	closesocket(socket);
}
#else
typedef int SOCKET;
typedef unsigned short WORD;
typedef unsigned int DWORD;
typedef hostent HOSTENT;
typedef sockaddr_in SOCKADDR_IN;
#define SOCKET_ERROR -1
#define INVALID_SOCKET -1
#endif


const int SIZE_OF_BUFFER_LEN_TYPE = sizeof(unsigned short);

//http://localhost/mu/server.php
//const char* init_worldUrl = "http://www.seamless3d.com/wrl/test_mu/test_mu.php";

float getFloat(const M_<char>& buf, int& mark){
	float val = *reinterpret_cast<const float*>(&buf[mark]);
	mark+=sizeof(float);
	return val;
}
Rotation getRotation(const M_<char>& buf, int& mark){
	Rotation val = *reinterpret_cast<const Rotation*>(&buf[mark]);
	mark+=sizeof(Rotation);
	return val;
}
Vec3f getVec3f(const M_<char>& buf, int& mark){
	Vec3f val = *reinterpret_cast<const Vec3f*>(&buf[mark]);
	mark+=sizeof(Vec3f);
	return val;
}
bool getBool(const M_<char>& buf, int& mark){
	bool val = *reinterpret_cast<const bool*>(&buf[mark]);
	mark+=sizeof(bool);
	return val;
}
enum ChatInstructions{
	CLIENT_REQUEST_TO_JOIN=1,
	SERVER_REPLY_FOR_JOIN,
	SERVER_REPLY_NEED_HIGHER_VERSION,
	NEW_CLIENT_JOINED,
	CLIENT_LEFT,
	CHAT_FROM_CLIENT,
	MESSAGE_FROM_SERVER,
	CLIENT_ORIENTATION_POSITION,
	CLIENT_NAV_KEY,
	CLIENT_GESTURE,
	TOUCH_SENSOR,

	CLIENT_LEFT_KEY=1,
	CLIENT_RIGHT_KEY=2,
	CLIENT_BACKWARD_KEY=4,
	CLIENT_FORWARD_KEY=8,
};
class ServerReplyForJoin{
	char instruction;
	unsigned short length;
	short int id;
	//any number of
	short int _id;
	String name;
	String avatarUrl;
	bool male;
	String voice;
	Rotation orientation;
	Vec3f position;
};
class NewClientJoined{
	char instruction;
	unsigned short length;
	short int id;
	short int numOfOtherClients;
	String name;
	String avatarUrl;
	bool male;
	String voice;
};
class ClientLeft{
	char instruction;
	unsigned short length;
	short int id;
};
class ChatFromClient{
	char instruction;
	unsigned short length;
	short int id;
	String message;
};
class MessageFromServer{
	char instruction;
	unsigned short length;
	String message;
};
class ClientOrientationPosition{
	char instruction;
	unsigned short length;
	short int id;
	Rotation orientation;
	Vec3f position;
};
class Touch_Sensor{
	char instruction;
	unsigned short length;
	short int id; //who clicked
	short int who_got_clicked_id;
	String nodeName;
};
class ClientNavKey{
	char instruction;
	unsigned short length;
	short int id;
	char keys;	//1=left 2=right 4=backward; 8=forward
};

void dLog(String str){
	FILE* stream;
	puts(str.c_str());
	str+="\n";
	stream = fopen("serverLog.txt","a+t");
	fputs(str.c_str(), stream);
	fclose(stream);
}
void formatTime(String& str, const time_t& rawtime){
	char chars[100];
	tm* ptm = gmtime(&rawtime);
	//sprintf(chars, "GMT: %.2i/%.2i/%i %.2i:%.2i", ptm->tm_mday, ptm->tm_mon+1, ptm->tm_year+1900, (ptm->tm_hour)%24, ptm->tm_min);
	sprintf(chars, "GMT: %i/%.2i/%.2i %.2i:%.2i", ptm->tm_year+1900, ptm->tm_mon+1, ptm->tm_mday, (ptm->tm_hour)%24, ptm->tm_min);
	str+=chars;
}
void gmtTime(String& str){
	time_t rawtime;
	time(&rawtime);

	formatTime(str, rawtime);
}
fd_set g_readfds;
//fd_set g_writefds;
fd_set g_exceptfds;

fd_set g_t_readfds;
//fd_set g_t_writefds;
fd_set g_t_exceptfds;


class World;
class AvatarSocket : public Node{
	public:
typedef bool (AvatarSocket::*MultiTask)();

	MultiTask multiTaskA;
	MultiTask multiTaskB;

	Rotation orientation;
	Vec3f position;
	String nickname;
	String avatarUrl;
	bool male;
	String voice;
	short id;

	AvatarSocket():
		orientation(Rotation(0,1,0,0)),
		position(Vec3f(0,0,0)),
		nickname(String()),
		avatarUrl(String()),
		male(false),
		voice(String()),
		id(0),

		/////
		hInternetOpenUrl(0),
		socket(0),
		dwThrdParam(1),
		world(0),
		multiTaskA(&AvatarSocket::initSocket),
		multiTaskB(&AvatarSocket::socketToBuf0),
		limit(0),
		targetLen(0)
	{
//		multiTaskA = &AvatarSocket::decodeChatInstruction;
		//(this->*multiTask)();
	}
	virtual ~AvatarSocket()
	{
	};
	String chat;
	int hInternetOpenUrl;
	M_<char> internetBuf;
	bool decodeChatInstruction();
	bool decodeHttpInstruction();
	bool removeClient();
	bool initSocket();
	inline bool initSocketBase(SOCKET& socket, int& entryCount, String& str);
	inline bool decodeChatInstructionBase();

	bool getCheckID(const M_<char>& inBuf, int& mark, const int recvLen)const;

	bool getHTTP();
	bool socketToBuf0();
	bool socketToBuf1();
	bool socketToBuf_GET_HTTP0();

	bool deleteClientNode(M_<Ptr<AvatarSocket> >& m_client);
	bool deleteClient(M_<Ptr<AvatarSocket> >& m_client);
	bool deleteClient(M_<Ptr<World> >& m_world);

	SOCKET socket;

	DWORD dwThrdParam;

	M_<char> inBuf;
	M_<char> outBuf;
	World* world;

private:
	static char stackBuf[1024];
	static String str;
	static String str2;
	int limit;
	int targetLen;
};

char AvatarSocket::stackBuf[1024];
String AvatarSocket::str;
String AvatarSocket::str2;





typedef  Vec3f (Rotation::*RotationMultiplyRotationFP)()const; 

void abc(){

	RotationMultiplyRotationFP rotationMultiplyRotationFP = &Rotation::getAxis;
	Rotation rot;
	AvatarSocket as;

	typedef bool (AvatarSocket::*fred)();

}
class World : public Node{
	public:
	String url;
	String title;
	String nickCheckedOut;
	time_t timeCheckedOut;
	M_<Ptr<AvatarSocket> > m_avatarSocket;
	void sendToAll(M_<char>& buf);
	void sendToAll(const String chatStr);
	void sendToAllExcept(M_<char>& buf, AvatarSocket* block);
	virtual ~World()
	{
	};
};

M_<Ptr<World> > g_m_world;
M_<Ptr<AvatarSocket> > g_m_clientInit;

String g_serverUrl;


void showError(const char* s){
#ifdef WIN32
	int error=WSAGetLastError();
	String szError=s;
	szError+=" ";
	if(error==WSANOTINITIALISED)szError+="WSANOTINITIALISED";
	else if(error==WSAENETDOWN)szError+="WSAENETDOWN";
	else if(error==WSAEACCES)szError+="WSAEACCES";
	else if(error==WSAEINTR)szError+="WSAEINTR";
	else if(error==WSAEINPROGRESS)szError+="WSAEINPROGRESS";
	else if(error==WSAEFAULT)szError+="WSAEFAULT";
	else if(error==WSAENETRESET)szError+="WSAENETRESET";
	else if(error==WSAENOBUFS)szError+="WSAENOBUFS";
	else if(error==WSAENOTCONN)szError+="WSAENOTCONN";
	else if(error==WSAENOTSOCK)szError+="WSAENOTSOCK";
	else if(error==WSAEOPNOTSUPP)szError+="WSAEOPNOTSUPP";
	else if(error==WSAESHUTDOWN)szError+="WSAESHUTDOWN";
	else if(error==WSAEWOULDBLOCK)szError+="WSAEWOULDBLOCK";
	else if(error==WSAEMSGSIZE)szError+="WSAEMSGSIZE";
	else if(error==WSAEHOSTUNREACH)szError+="WSAEHOSTUNREACH";
	else if(error==WSAEINVAL)szError+="WSAEINVAL";
	else if(error==WSAECONNABORTED)szError+="WSAECONNABORTED";
	else if(error==WSAECONNRESET)szError+="WSAECONNRESET";
	else if(error==WSAETIMEDOUT)szError+="WSAETIMEDOUT";

	

	dLog(szError.c_str());

	WSASetLastError (0);
#else
	String szError=s;
	szError+=strerror(errno);
	dLog(szError.c_str());
#endif 
}
int advancePastUrlPrefix(String& url){
	if(url.length < 11){
		return 0;
	}
	const char* p = url.c_str();

	if(	p[0] == 'h' &&
		p[1] == 't' &&
		p[2] == 't' &&
		p[3] == 'p' &&
		p[4] == ':' &&
		p[5] == '/' &&
		p[6] == '/')
	{
		if(p[7] == 'w' && p[8] == 'w' && p[9] == 'w' && p[10] == '.'){
			return 11;
		}
		else{
			return 7;
		}
	}
	return 0;
}
bool compareUrl(String& urlA, String& urlB){
	int idxA = advancePastUrlPrefix(urlA);
	int idxB = advancePastUrlPrefix(urlB);

	int lenA = urlA.length - idxA;
	int lenB = urlB.length - idxB;

	if(lenA != lenB){
		return false;
	}
	while(idxA < lenA){
		if(urlA[idxA++] != urlB[idxB++]){
			return false;
		}
	}
	return true;
}

bool joinPHP(String filename, String& htmlStr, String& ip){
	SOCKET itvSocket;
    HOSTENT *hostEntry;

	String domain;
	{	
		int i = filename.indexOf("/") + 2;
		domain = filename.substring(i);
		i = domain.indexOf("/") - 1;
		domain = domain.substring(0, i);
	}
	hostEntry = gethostbyname(domain.c_str());
#ifdef WIN32
	if (WSAGetLastError() != 0) {
        return false;
	}
#endif
	itvSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (itvSocket == INVALID_SOCKET) {
		return false;
	}
	
	SOCKADDR_IN saServer;
	saServer.sin_family = AF_INET;
	//saServer.sin_addr = *((LPIN_ADDR)*hostEntry->h_addr_list);
	saServer.sin_addr.s_addr = *(unsigned long*)hostEntry->h_addr_list[0];
	saServer.sin_port = htons(80);	// html port
	
	// connect to the server
	if(SOCKET_ERROR == connect(itvSocket, (sockaddr *)&saServer, sizeof(sockaddr))){
		dLog("connect(itvSocket, (LPSOCKADDR)&saServer, sizeof(saServer)) ");
		close(itvSocket);
		return false;
	}

	String str=String("GET ");
	str+=filename;
	//str+=" HTTP/1.1\r\nHost: ";
	str+=" HTTP/1.1\r\n\
Host: ";
	str+=domain;
	str += "\r\nUser-Agent: Seamless3d";

	str+="\r\n\r\n\r\n";
	if( SOCKET_ERROR == send(itvSocket, str.c_str(), str.length, 0)){
		dLog("send(itvSocket, str.c_str(), str.length, 0)");
		close(itvSocket);	
		return false;
	}

	char szBuf[2048];

	int txtLength = recv(itvSocket, szBuf, sizeof(szBuf)-1, 0);

	if (txtLength == INVALID_SOCKET)	{
		dLog("recv() INVALID_SOCKET");
		close(itvSocket);
		return false;
	}
	else if (txtLength == 0)	{
		dLog("recv() 0");
		close(itvSocket);
		return false;
	}
	szBuf[txtLength]=0;
	ip=szBuf;

	close(itvSocket);

	bool chunked = false;
	if(ip.indexOf("Transfer-Encoding: chunked") >= 0){
		chunked = true;
	}

	int i = ip.indexOf("\r\n\r\n");
	ip.stb(i);
	if(chunked){
		ip.stw(i);
		ip.stb(i);
	}
	ip=ip.substring(i);
	i = ip.indexOf("\r\n");
	if(i > 0){
		ip = ip.substring(0, i - 1);
	}

	i = filename.indexOf(".php");
	htmlStr = filename.substring(0, i);
	htmlStr += "smls";


	return true;
}

bool AvatarSocket::deleteClientNode(M_<Ptr<AvatarSocket> >& m_client){
	for(int k = m_client.length; --k>=0;){
		if(m_client[k] == this){
			m_client.deleteElement(k);
			return true;
		}
	}
	return false;
}
bool AvatarSocket::deleteClient(M_<Ptr<AvatarSocket> >& m_client){
	if(deleteClientNode(m_client)){
		g_numOfClients--;
		FD_CLR(socket, &g_readfds);
		FD_CLR(socket, &g_exceptfds);
		close(socket);
		gmtTime(str="\n");
		str+=" deleteClient ";
		str+=nickname;
		str+=" g_numOfClients = ";
		str+=g_numOfClients;
		dLog(str);
		return true;
	}
	return false;
}
bool AvatarSocket::deleteClient(M_<Ptr<World> >& m_world){
	for(int k = m_world.length; --k>=0;){
		if(deleteClient(g_m_world[k]->m_avatarSocket)){
			return true;
		}
	}
	return false;
}
inline bool AvatarSocket::initSocketBase(SOCKET& socket, int& entryCount, String& str){
	this->socket = socket;
	g_m_clientInit+=this;

#ifdef WIN32
	unsigned long iMode = 1;
	ioctlsocket(socket, FIONBIO, &iMode);
#else
	int flags = fcntl(socket,F_GETFL,0);
	fcntl(socket, F_SETFL, flags | O_NONBLOCK);
#endif

	FD_SET(socket, &g_readfds);
	FD_SET(socket, &g_exceptfds);
	id = ++entryCount;
	//clientInitList+=client;
	gmtTime(str="\n");
	str+=" create client g_numOfClients = ";
	str+=g_numOfClients;
	dLog(str);
	//multiTaskA = &AvatarSocket::initSocket;
	//(this->*multiTaskA)();
	//if(false == (this->*multiTaskA)()){
	//	if(deleteClient(g_m_clientInit)){
	//		dLog("failed to initialise client at initSocketBase");
	//	}
	//	else{
	//		dLog("failed to delete client at initSocketBase !!!");
	//		throw;
	//	}
	//}
	return true;
}
inline bool AvatarSocket::decodeChatInstructionBase(){
	if(FD_ISSET(socket, &g_t_exceptfds)){
		dLog("g_t_exceptfds at decodeChatInstructionBase");
		if(false == deleteClient(g_m_clientInit)){
			if(deleteClient(g_m_world)){
				dLog("failed to decode client at decodeChatInstructionBase");
			}
			else{
				dLog("failed to delete client at decodeChatInstructionBase!!!");
			}
		}
	}
	else if(FD_ISSET(socket, &g_t_readfds)){
		if(false == (this->*multiTaskA)()){
			if(false == deleteClient(g_m_clientInit)){
				if(deleteClient(g_m_world)){
					dLog("failed to initialise client at decodeChatInstructionBase2");
				}
				else{
					dLog("failed to delete client at decodeChatInstructionBase2!!!");
				}
			}
		}
	}
	return true;
}
void serverMain(short port){

	// Create a TCP/IP stream socket to "listen" with
	SOCKET	listenSocket = socket(AF_INET, SOCK_STREAM,	IPPROTO_TCP	);
	if(listenSocket == INVALID_SOCKET)	{
		showError("socket()");
		return;
	}
	int optVal = 1;
	if(setsockopt(listenSocket, SOL_SOCKET, SO_REUSEADDR, (char*)&optVal, sizeof(optVal)) < 0){
		showError("setsockopt()");
		return;
	}

	// Fill in the address structure
	SOCKADDR_IN saServer;
#ifndef WIN32
	bzero((char *) &saServer, sizeof(saServer));
#endif
	saServer.sin_family = AF_INET;
	saServer.sin_addr.s_addr = INADDR_ANY;
	saServer.sin_port = htons(port);

	// bind listen socket to port and all addresses
	if(bind(listenSocket, (struct sockaddr*)&saServer, sizeof(saServer)) < 0){
		showError("bind()");
		close(listenSocket);
		return;
	}


	// Set the socket to listen mode

	if(listen(listenSocket, SOMAXCONN) == SOCKET_ERROR){ // Number of connection request queue
		showError("listen()");
		close(listenSocket);
		return;
	}


	// Wait for an incoming request

	String str;
	int entryCount=0;
	FD_ZERO(&g_readfds);
	//FD_ZERO(&g_writefds);
	FD_ZERO(&g_exceptfds);
	FD_SET(listenSocket, &g_readfds);
	//FD_SET(listenSocket, &g_writefds);
	FD_SET(listenSocket, &g_exceptfds);

	while(1){
		g_t_readfds = g_readfds;
		//g_t_writefds = g_writefds;
		g_t_exceptfds = g_exceptfds;
		if(select(FD_SETSIZE, &g_t_readfds, 0, &g_t_exceptfds, 0) < 1){
			showError("select()");
			return;
		}
		for(int k = g_m_world.length; --k>=0;){
			World* world = g_m_world[k];
			M_<Ptr<AvatarSocket> >& m_avatarSocket = world->m_avatarSocket;
			for(int k = m_avatarSocket.length; --k>=0;){
				m_avatarSocket[k]->decodeChatInstructionBase();
			}
		}
		for(int k = g_m_clientInit.length; --k>=0;){
			g_m_clientInit[k]->decodeChatInstructionBase();
		}
		if(FD_ISSET(listenSocket, &g_t_exceptfds)){
			dLog("listenSocket g_t_exceptfds");
		}
		else if(FD_ISSET(listenSocket, &g_t_readfds)){
			SOCKET sock = accept(listenSocket, 0, 0);

			if(sock == INVALID_SOCKET){
				dLog("listenSocket == INVALID_SOCKET");
			}
			else {
				if(++g_numOfClients >= MAX_CLIENT){
					dLog("maximum number of clients exceeded!!!");
					return;
				}
				Ptr<AvatarSocket> client = new AvatarSocket();
				client->initSocketBase(sock, entryCount, str);
			}
		}
	}
	// Close all sockets before exiting
	for(int i=g_m_world.length; --i>=0;){
		World* world = g_m_world[i];
		M_<Ptr<AvatarSocket> >& m_avatarSocket = world->m_avatarSocket;
		for(int i=m_avatarSocket.length; --i>=0;){ //delete our pointer from the list list
			close(m_avatarSocket[i]->socket);
		}
	}
	close(listenSocket);

	//Not sure this is needed but does no harm
	
	return;
}

bool joinSession(String& ip){

	//	printf("Client connecting to server: %s on port: %d",
	//				szServer, nPort);
	
	// Find the server
    HOSTENT *hostEntry;
	in_addr		iaHost;
	
	iaHost.s_addr = inet_addr(ip.c_str());
	
	hostEntry = gethostbyaddr((const char *)&iaHost, sizeof(struct in_addr), AF_INET);
	
    if (hostEntry == NULL) {
        return false;
    }
	
	// Create a TCP/IP stream socket
	
	SOCKET chatSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (chatSocket == INVALID_SOCKET) {
		return false;
	}
	
	// Fill in the address structure
	SOCKADDR_IN saServer;
	
	saServer.sin_family = AF_INET;
	//saServer.sin_addr = *((LPIN_ADDR)*hostEntry->h_addr_list);
	saServer.sin_addr.s_addr = *(unsigned long*)hostEntry->h_addr_list[0];
	
	
	saServer.sin_port = htons(PORT);
	
	// connect to the server
	
	//if(SOCKET_ERROR == connect(chatSocket, (LPSOCKADDR)&saServer, sizeof(struct sockaddr))){
	if(SOCKET_ERROR == connect(chatSocket, (sockaddr *)&saServer, sizeof(sockaddr))){
		close(chatSocket);
		return false;
	}
	dLog("World already served by " + ip);
	printf("Press any key to exit");

	close(chatSocket);

	getchar();
	

	return true;
}


int main(int argc, char* argv[]){
	printf("Seamless3d Server %.3f\n", g_version);
#ifdef WIN32
	WORD wVersionRequested = MAKEWORD(2,2);
	WSADATA wsaData;

	// Initialize WinSock and check version
	WSAStartup(wVersionRequested, &wsaData);
	if (wsaData.wVersion != wVersionRequested){	
		dLog("\n Wrong version\n");
		return 0;
	}
#endif
	if(argc == 1){
		dLog(String("Port:\t") + PORT);
		serverMain(PORT);
#ifdef WIN32
		WSACleanup();
#endif
		return 0;
	}

	String htmlStr;
	String ip;
	g_serverUrl = argv[1];
	if(g_serverUrl[g_serverUrl.length-1] != '/'){
		g_serverUrl+='/';
	}
	String str;
	gmtTime(str="");
	dLog(str);
	
	String requestUrl = g_serverUrl;
	requestUrl += "smls_server.php?session=join";
	dLog(requestUrl);
	joinPHP(requestUrl, htmlStr, ip);

	if(false == joinSession(ip)){

		requestUrl = g_serverUrl;
		requestUrl += "smls_server.php?session=create";
		
		if(joinPHP(requestUrl, htmlStr, ip)){
			gmtTime(str="");
			dLog(str);
			dLog(requestUrl);
			dLog("IP:\t" + ip);
			dLog(String("Port:\t") + PORT);
			serverMain(PORT);
		}
	}

#ifdef WIN32
	WSACleanup();
#endif
}
void insertBool(M_<char>& buf, bool data){
	buf+=static_cast<char>(data);
}
void insertShort(M_<char>& buf, short data){
	for(int i=2; --i>=0;){
		buf+=static_cast<char>(data);
		data>>=8;
	}
}
void insertUnsignedShort(M_<char>& buf, int mark, int data){
	if(data>=0x10000){
		dLog("*insertUnsignedShort excedes 16 bits*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
		return;
	}
	for(int i=2; --i>=0;){
		buf[mark++]=static_cast<char>(data);
		data>>=8;
	}
}
void insertInt32(M_<char>& buf, int data){
	for(int i=4; --i>=0;){
		buf+=static_cast<char>(data);
		data>>=8;
	}
}
void insertFloat(M_<char>& buf, float data){
	insertInt32(buf, *reinterpret_cast<int*>(&data));
}
void insertRotation(M_<char>& buf, Rotation data){
	insertInt32(buf, *reinterpret_cast<int*>(&data.x));
	insertInt32(buf, *reinterpret_cast<int*>(&data.y));
	insertInt32(buf, *reinterpret_cast<int*>(&data.z));
	insertInt32(buf, *reinterpret_cast<int*>(&data.angle));
}
void insertVec3f(M_<char>& buf, Vec3f data){
	insertInt32(buf, *reinterpret_cast<int*>(&data.x));
	insertInt32(buf, *reinterpret_cast<int*>(&data.y));
	insertInt32(buf, *reinterpret_cast<int*>(&data.z));
}
short int getShort(const M_<char>& buf, int& mark){
	short int val = *reinterpret_cast<const short int*>(&buf[mark]);
	mark+=2;
	return val;
}
unsigned short int getUnsignedShort(const M_<char>& buf, int& mark){
	short int val = *reinterpret_cast<const short int*>(&buf[mark]);
	mark+=2;
	return val;
}
void addStr(String& str, const M_<char>& buf, int& mark){
	int i=str.length;
	str+=&buf.getArray()[mark];
	mark+=str.length - i + 1;
}
void insertStr(M_<char>& buf, const String str){
	for(int i=0; i<str.length; i++){
		buf+=str[i];
	}
	buf+=0;
}
void insertStr(M_<char>& outBuf, M_<char>& inBuf, int& mark){
	do{
		outBuf+=inBuf[mark];
	}while (inBuf[mark++]!=0);
}
void World::sendToAll(M_<char>& buf){
	insertUnsignedShort(buf, 1, buf.length);
	for(int i=m_avatarSocket.length;--i>=0;){
		if(send(m_avatarSocket[i]->socket, buf.getArray(), buf.length, 0) == SOCKET_ERROR){
			showError((m_avatarSocket[i]->nickname + "send() = SOCKET_ERROR & sendToAll").c_str());
			//closesocket(g_m_avatarSocket[i]->socket);	
		}
	}
}
void insertInstruction(M_<char>& buf, char ChatInstructions){
	buf=ChatInstructions;
	buf.length+=SIZE_OF_BUFFER_LEN_TYPE;
}
void insertInstruction(M_<char>& buf, char ChatInstructions, short id){
	insertInstruction(buf, ChatInstructions);
	insertShort(buf, id);
}

void World::sendToAllExcept(M_<char>& buf, AvatarSocket* block){
	insertUnsignedShort(buf, 1, buf.length);
	for(int i=m_avatarSocket.length;--i>=0;){
		if(m_avatarSocket[i]==block){
			continue;
		}
		if(send(m_avatarSocket[i]->socket, buf.getArray(), buf.length, 0) == SOCKET_ERROR){
			showError((m_avatarSocket[i]->nickname + " send() = SOCKET_ERROR").c_str());
		}
	}
}

void sendToOne(M_<char>& buf, AvatarSocket* one){
	insertUnsignedShort(buf, 1, buf.length);
	if(send(one->socket, buf.getArray(), buf.length, 0) == SOCKET_ERROR){
		showError("send() = SOCKET_ERROR");
	}
}
void sendToOneHtml(M_<char>& buf, AvatarSocket* one){
	if(send(one->socket, buf.getArray(), buf.length, 0) == SOCKET_ERROR){
		showError("send() = SOCKET_ERROR");
	}
}


void World::sendToAll(const String chatStr){
	static M_<char> buf;
	buf.length=0;

	insertInstruction(buf, MESSAGE_FROM_SERVER);

	insertStr(buf, chatStr);

	sendToAll(buf);
}
bool AvatarSocket::getCheckID(const M_<char>& inBuf, int& mark, const int recvLen)const{
	if(mark+2 > recvLen){
		mark = recvLen;
		return false;
	}
	if(id != getShort(inBuf, mark)){
		mark = recvLen;
		return false;
	}
	return true;
}
bool checkThereAreEnoughBytes(const M_<char>& inBuf, int& mark, const int typeSize){
	if(mark + typeSize <= inBuf.length){
		return true;
	}
	return false;
}
bool checkThereAreEnoughBytes(const M_<char>& inBuf, int& mark, const int typeSize, const char* msg){
	if(checkThereAreEnoughBytes(inBuf, mark, typeSize)){
		return true;
	}
	String str;
	gmtTime(str="");
	str+=" Error: ";
	str+=msg;
	str+=" not enough bytes for variable type !!!";
	dLog(str);
	return false;
}
bool checkThereAreEnoughBytesBool(const M_<char>& inBuf, int& mark, const char* msg, bool& var){
	if(checkThereAreEnoughBytes(inBuf, mark, sizeof(var))){
		var = getBool(inBuf, mark);
		return true;
	}
	return false;
}
bool checkThereAreEnoughBytesRotation(const M_<char>& inBuf, int& mark, const char* msg, Rotation& var){
	if(checkThereAreEnoughBytes(inBuf, mark, sizeof(var))){
		var = getRotation(inBuf, mark);
		return true;
	}
	return false;
}
bool checkThereAreEnoughBytesVec3f(const M_<char>& inBuf, int& mark, const char* msg, Vec3f& var){
	if(checkThereAreEnoughBytes(inBuf, mark, sizeof(var))){
		var = getVec3f(inBuf, mark);
		return true;
	}
	return false;
}
const char* checkStringIsLegit(const M_<char>& inBuf, int& mark){
	int remMark = mark;
	const char* p = inBuf.getArray();
	const int inBufLen = inBuf.length;
	for(; mark < inBufLen; mark++){
		if(p[mark]==0){
			mark++;
			return p + remMark;
		}
	}
	return 0;
}
bool checkStringIsLegit(const M_<char>& inBuf, int& mark, const char* msg, String& str){
	const char* charPtr = checkStringIsLegit(inBuf, mark);
	if(charPtr){
		str=charPtr;
		return true;
	}
	gmtTime(str="");
	str+=" Error: ";
	str+=msg;
	str+=" not legit string !!!";
	dLog(str);
	return false;
}
void specifyHowLongAgo(String& str, const time_t& currentTime, const time_t& since){
	int seconds = currentTime - since;
	int minutes = seconds/60;
	int hours = minutes/60;
	int days = hours/24;
	if(seconds < 60){
		str+=seconds;
		str+=" seconds ago";
	}
	else if(minutes < 60){
		if(minutes==1){
			str+="a minute";
		}
		else{
			str+=minutes;
			str+=" minutes";
		}
		int remainder = seconds - minutes * 60;
		if(remainder==1){
			str+=" and a second ago";
		}
		else{
			str+=" and ";
			str+=remainder;
			str+=" seconds ago";
		}
	}
	else if(hours < 24){
		if(hours==1){
			str+="an hour";
		}
		else{
			str+=hours;
			str+=" hours";
		}

		int remainder = minutes - hours * 60;
		if(remainder==1){
			str+=" and a minute ago";
		}
		else{
			str+=" and ";
			str+=remainder;
			str+=" minutes ago";
		}
	}
	else {
		if(days==1){
			str+="a day";
		}
		else{
			str+=days;
			str+=" days";
		}
		int remainder = hours - days * 24;
		if(remainder==1){
			str+=" and an hour ago";
		}
		else{
			str+=" and ";
			str+=remainder;
			str+=" hours ago";
		}
	}
}
void moveWorldToFirstInList(World* world){
	for(int k = g_m_world.length; --k>0;){
		if(g_m_world[k] == world){
			while(--k>=0){
				g_m_world[k + 1] = g_m_world[k];
			}
		}
	}
	g_m_world[0] = world;
}
bool AvatarSocket::socketToBuf0(){
	targetLen = 0;
	const int initialLen = 3;
	inBuf.length=0;
	int recvLen = recv(socket, stackBuf, initialLen, 0);

	if(recvLen == INVALID_SOCKET){
		dLog(nickname + " socketToBuf0 recv() INVALID_SOCKET");
		return false;
	}
	else if(recvLen == 0)	{
		dLog(nickname + " socketToBuf0 recv() 0");
		return false;
	}
	for(int i = 0; i < recvLen; i++){
		inBuf+=stackBuf[i];
	}
	if(inBuf[0]>='A'){//html
		multiTaskA = &AvatarSocket::getHTTP;
		multiTaskB = &AvatarSocket::socketToBuf_GET_HTTP0;
		return true;;
	}
	else{
		int mark = 1;
		targetLen = getUnsignedShort(inBuf, mark);
	}

	limit = targetLen - inBuf.length;
	if(limit >= sizeof(stackBuf)){
		limit = sizeof(stackBuf);
	}
	multiTaskB = &AvatarSocket::socketToBuf1;
	return true;
}
bool AvatarSocket::socketToBuf1(){
	int recvLen = recv(socket, stackBuf, limit, 0);
	if(recvLen == INVALID_SOCKET)	{
		dLog(nickname + " socketToBuf1 recv() INVALID_SOCKET");
		return false;
	}
	else if(recvLen == 0)	{
		dLog(nickname + " socketToBuf1 recv() 0");
		return false;
	}

	for(int i = 0; i < recvLen; i++){
		inBuf+=stackBuf[i];
	}
	return true;
}
bool AvatarSocket::socketToBuf_GET_HTTP0(){
	int recvLen = recv(socket, stackBuf, sizeof(stackBuf), 0);
	if(recvLen == INVALID_SOCKET){
		dLog(nickname + " socketToBuf_GET_HTTP0 recv() INVALID_SOCKET");
		return false;
	}
	else if(recvLen == 0)	{
		dLog(nickname + " socketToBuf_GET_HTTP0 recv() 0");
		return false;
	}
	for(int i = 0; i < recvLen; i++){
		inBuf+=stackBuf[i];
	}
	return true;
}
bool endOfHTTPMessage(M_<char>& inBuf){ 
	for(int i = 0; i<inBuf.length-3; i++){
		if( inBuf[i+0] == '\r' &&
			inBuf[i+1] == '\n' &&
			inBuf[i+2] == '\r' &&
			inBuf[i+3] == '\n'
		){
			return true;
		}
	}
	return false;
}
bool AvatarSocket::getHTTP(){

	bool socketToBufFlag = (this->*multiTaskB)();

	if(socketToBufFlag == false){
		deleteClient(g_m_clientInit);
		dLog("HTML Browser logged out");
		return true;
	}
	if(endOfHTTPMessage(inBuf)){

		dLog("HTML Browser logged in");
		inBuf+=0;
		String str = inBuf.getArray();

		String content;

		time_t currentTime;
		time(&currentTime);

		for(int k = 0; k < g_m_world.length; k++){
			World* world = g_m_world[k];
			content += "<a href=\"";
			content += world->url;
			content += "\"><img alt=\"\" src=\"";

			int dotIdx = world->url.length;
			while(--dotIdx >= 0){
				if(world->url[dotIdx]=='.'){
					break;
				}
			}
			if(dotIdx>=0){
				content += world->url.substring(0, dotIdx - 1);
			}
			content += "_tn.jpg\" /></a><b><a href=\"";

			//content += g_serverUrl;
			//content += "?url=";
			content += world->url;
			/*
			content += "&title=";
			for(int i = 0; i < world->title.length; i++){
				char chr = world->title[i];
				if(chr==' '){
					chr = '%';
				}
				content += chr;
			}
			*/
			content += "\"> ";
			content += world->title;
			M_<Ptr<AvatarSocket> >& m_avatarSocket = world->m_avatarSocket;
			if(m_avatarSocket.length){
				content += "</a></b> Occupants: ";
				for(int i=m_avatarSocket.length; --i>=0;){
					if(m_avatarSocket[i]!=this){
						AvatarSocket* client = m_avatarSocket[i];
						content += client->nickname;
						if(i == 0){
							break;
						}
						content += " - ";
					}
				}
			}
			else{
				content += "</a></b> ";
				content += world->nickCheckedOut;
				content += " checked out ";
				specifyHowLongAgo(content, currentTime, world->timeCheckedOut);
			}
			content += "<br/>";
		}
		if(g_m_world.length==0){
			content = "The Seamless3d Server is running but no worlds are open.";
		}

		//tmpStr = "HTTP/1.1 200 OK\r\n"
		//	"Date: Sat, 24 May 2008 08:54:37 GMT\r\n"
		//	"Server: Apache\r\n"
		//	"X-Powered-By: PHP/4.4.8\r\n"
		//	"Vary: Host\r\n"
		//	"Keep-Alive: timeout=15, max=100\r\n"
		//	"Connection: Keep-Alive\r\n"
		//	"Transfer-Encoding: chunked\r\n"
		//	"Content-Type: text/html\r\n"
		//	"\r\n"
		//	"34\r\n" 
		//	"http://www.seamless3d.com/wrl/test_mu/test_mu.smls"
		//	"\r\nd\r\n"
		//	"144.138.47.13\r\n"
		//	"\r\n0\r\n\r\n";

		str = "HTTP/1.1 200 OK\r\n"
			"Server: Seamless3d " + String(g_version) + "\r\n"
			"Connection: Close\r\n"
			"Content-Length: " + String(content.length) + "\r\n"
			"Content-Type: text/html\r\n\r\n" + content;


		insertStr(outBuf, str);
		//sendToOne(outBuf, this);
		sendToOneHtml(outBuf, this);
	}
	return true;
}
bool AvatarSocket::initSocket(){

	bool socketToBufFlag = (this->*multiTaskB)();

	if(socketToBufFlag == false){
		close(socket);
		return false;
	}
	if(targetLen > 0 && targetLen == inBuf.length){
		multiTaskB = &AvatarSocket::socketToBuf0;
		if(inBuf[0]==CLIENT_REQUEST_TO_JOIN){
			multiTaskA = &AvatarSocket::decodeChatInstruction;
			int mark = 1 + SIZE_OF_BUFFER_LEN_TYPE;

			if(getFloat(inBuf, mark) < 2.146-.001){
				insertInstruction(outBuf, SERVER_REPLY_NEED_HIGHER_VERSION, id);
				insertFloat(outBuf, g_version);
				sendToOne(outBuf, this);
				close(socket);
				return false;
			}
			
			if(!checkStringIsLegit(inBuf, mark, "CLIENT_REQUEST_TO_JOIN world url", str)){
				return false;
			}
			String title;
			if(!checkStringIsLegit(inBuf, mark, "CLIENT_REQUEST_TO_JOIN title", title)){
				return false;
			}

			for(int k = g_m_world.length; --k>=0;){
				if(compareUrl(g_m_world[k]->url, str)){
					g_m_world[k]->title = title; //update title in case it has changed
					world = g_m_world[k];
					moveWorldToFirstInList(world);
					break;
				}
			}
			if(world == 0){
				world = new World;
				world->url = str;
				world->title = title;
				g_m_world.insertElement(0, world);
				dLog("new world: " + world->title + " " + world->url);
			}

			if(false == checkStringIsLegit(inBuf, mark, "CLIENT_REQUEST_TO_JOIN nickname", nickname)){
				return false;
			}
			if(false == checkStringIsLegit(inBuf, mark, "CLIENT_REQUEST_TO_JOIN avatarUrl", avatarUrl)){
				return false;
			}
			if(false == checkThereAreEnoughBytesBool(inBuf, mark, "CLIENT_REQUEST_TO_JOIN male", male)){
				return false;
			}
			if(false == checkStringIsLegit(inBuf, mark, "CLIENT_REQUEST_TO_JOIN voice", voice)){
				return false;
			}
			if(false == checkThereAreEnoughBytesRotation(inBuf, mark, "CLIENT_REQUEST_TO_JOIN orientation", orientation)){
				return false;
			}
			if(false == checkThereAreEnoughBytesVec3f(inBuf, mark, "CLIENT_REQUEST_TO_JOIN position", position)){
				return false;
			}

			M_<Ptr<AvatarSocket> >& m_avatarSocket = world->m_avatarSocket;
			m_avatarSocket+=this;

			deleteClientNode(g_m_clientInit);

			str="CLIENT_REQUEST_TO_JOIN ";
			str+=' ';
			str+=world->title;
			str+=' ';
			str+=static_cast<int>(id);
			str+=" : ";
			str+=nickname;
			str+="=";
			str+=avatarUrl;
			str+=" male=";
			str+=male;
			str+=" voice=";
			str+=voice;
			str+=" orientation=";
			str+=orientation.toString();
			str+=" position=";
			str+=position.toString();
			dLog(str);


			insertInstruction(outBuf, SERVER_REPLY_FOR_JOIN, id);
			insertShort(outBuf, m_avatarSocket.length-1);
			for(int i = m_avatarSocket.length; --i>=0;){
				if(m_avatarSocket[i]!=this){
					AvatarSocket& client=*m_avatarSocket[i];
					insertShort(outBuf, client.id);
					insertStr(outBuf, client.nickname);
					insertStr(outBuf, client.avatarUrl);
					insertBool(outBuf, client.male);
					insertStr(outBuf, client.voice);
					insertRotation(outBuf, client.orientation);
					insertVec3f(outBuf, client.position);
				}
			}
			sendToOne(outBuf, this);

			insertInstruction(outBuf, NEW_CLIENT_JOINED, id);
			insertStr(outBuf,nickname);
			insertStr(outBuf,avatarUrl);
			insertBool(outBuf, male);
			insertStr(outBuf,voice);
			insertRotation(outBuf, orientation);
			insertVec3f(outBuf, position);
			world->sendToAllExcept(outBuf, this);
		}
	}
	return true;
}
bool AvatarSocket::removeClient(){
	if(world){
		insertInstruction(outBuf, CLIENT_LEFT, id);
		
		//now get rid of this client from client list

		M_<Ptr<AvatarSocket> >& m_avatarSocket = world->m_avatarSocket;
		Ptr<World> world = this->world; //needed if our world gets deleted
		Ptr<AvatarSocket> avatarSocket = this; //needed for deleting us

		deleteClient(m_avatarSocket);

		world->nickCheckedOut=nickname;
		time(&world->timeCheckedOut);
		str=nickname;
		str+=" logged out ";
		str+=world->title;
		dLog(str.c_str());
		world->sendToAll(outBuf);
		
		str=world->title;
		str+=" m_avatarSocket.length ";
		str+=m_avatarSocket.length;
		dLog(str.c_str());

		/* remember empty worlds for 2.143
		if(m_avatarSocket.length == 0){
			for(int i=g_m_world.length; --i>=0;){
				if(g_m_world[i]==world){
					g_m_world.deleteElement(i);
					break;
				}
			}
		}
		*/
		if(m_avatarSocket.length==1){
			str = "you are alone in ";
			str += world->title;
			str += ' ';
			str += m_avatarSocket[0]->nickname;
			world->sendToAll(str);
		}
		else{
			str=m_avatarSocket.length;
			str+=" exist in ";
			str += world->title;
			world->sendToAll(str);
		}
		avatarSocket = 0;
		world = 0;
		//no member variables can be used from here on
	}

	return true;
}
bool AvatarSocket::decodeHttpInstruction(){
	bool socketToBufFlag = socketToBuf0();
	if(socketToBufFlag == false){
		return removeClient();
	}
	return true;
}

bool AvatarSocket::decodeChatInstruction(){
	outBuf.length=0;

	bool socketToBufFlag = (this->*multiTaskB)();

	if(socketToBufFlag == false){
		return removeClient();
	}
	if(targetLen > 0 && targetLen == inBuf.length){

		multiTaskB = &AvatarSocket::socketToBuf0;

		char instruction = inBuf[0];
		int mark = 1 + SIZE_OF_BUFFER_LEN_TYPE;
		switch(instruction){
			case CHAT_FROM_CLIENT:{
				insertInstruction(outBuf, CHAT_FROM_CLIENT, id);
				if(false == getCheckID(inBuf, mark, inBuf.length)){
					gmtTime(str="");
					str+=" CHAT_FROM_CLIENT ID missing or does match thread !!!";
					dLog(str);
					break;
				}
				if(!checkStringIsLegit(inBuf, mark, "CHAT_FROM_CLIENT", str2)){
					break;
				}
				str="CHAT_FROM_CLIENT ";
				str+=nickname;
				str+=" : ";
				str+=str2;
				dLog(str);
				insertStr(outBuf, str2);
				world->sendToAll(outBuf);
				if(str2 == "version?"){
					insertInstruction(outBuf, MESSAGE_FROM_SERVER);
					str2 = "server: version = ";
					str2 += g_version;
					insertStr(outBuf, str2);
					world->sendToAll(outBuf);
				}
				break;
			}
			case CLIENT_ORIENTATION_POSITION:{
				insertInstruction(outBuf, CLIENT_ORIENTATION_POSITION, getShort(inBuf, mark));
				
				orientation = getRotation(inBuf, mark);
				position = getVec3f(inBuf, mark);

				insertRotation(outBuf, orientation);
				insertVec3f(outBuf, position);

				str=nickname;
				str+=" position=";
				str+=position.toString();
				dLog(str);

				world->sendToAllExcept(outBuf, this);
				break;
			}
			case CLIENT_NAV_KEY:{
				insertInstruction(outBuf, CLIENT_NAV_KEY, getShort(inBuf, mark));
				
				const char keys=inBuf[mark++];
				if(keys&CLIENT_LEFT_KEY){
					orientation *= Rotation(0,1,0,PI*.01);
				}
				if(keys&CLIENT_RIGHT_KEY){
					orientation *= Rotation(0,1,0,-PI*.01);
				}
				if(keys&CLIENT_BACKWARD_KEY){
					position += orientation * Vec3f(0,0,-.05);
				}
				if(keys&CLIENT_FORWARD_KEY){
					position += orientation * Vec3f(0,0,.05);
				}
				outBuf+=keys;
				world->sendToAllExcept(outBuf, this);
				break;
			}
			case CLIENT_GESTURE:{
				insertInstruction(outBuf, CLIENT_GESTURE, getShort(inBuf, mark));
				
				int gestureIdx = getShort(inBuf, mark);

				insertShort(outBuf, gestureIdx);

				str=nickname;
				str+=" gesture=";
				str+=gestureIdx;
				dLog(str);

				world->sendToAllExcept(outBuf, this);
				break;
			}
			case TOUCH_SENSOR:{
				insertInstruction(outBuf, TOUCH_SENSOR, getShort(inBuf, mark));
				short whoGotClicked = getShort(inBuf, mark);
				if(!checkStringIsLegit(inBuf, mark, "TOUCH_SENSOR", str2)){
					break;
				}
				str="TOUCH_SENSOR ";
				str+=str2;
				str+=' ';
				str+=nickname;
				dLog(str);

				insertShort(outBuf, whoGotClicked);
				insertStr(outBuf, str2);
				world->sendToAllExcept(outBuf, this);
				break;
			}
			default:{
				dLog("received unknown instruction from client!");
				break;
			}
		}
	}
	return true;
}
//template <> void M_<Ptr<World> >::insertElement(int targetIndex, Ptr<World> value){iinsertElement(*this, targetIndex, value);}
