/* Seamless3d Server

Copyright (C) 2002-2009 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#pragma warning( disable : 4996 4244 4305)

#include "var.h"

char floatFormat[]={"%.6f"};

short ControlWord;
short ModifiedControlWord;


//lowLevelCode
float atof2(const char* str){
	long double num=0;
	int i=0;
	long double negFlag=1;
	if(str[i]=='-'){
		negFlag=-1;
		i++;
	}
	while(1){
		if(str[i]=='.'){
			i++;
			float counter=.1;
			while(str[i]>='0'&&str[i]<='9'){
				num+=counter*(float)(str[i]-'0');
				counter*=.1;
				i++;
			}
			if(str[i]=='E'||str[i]=='e'){
				num=0;
				negFlag=0;
			}
			break;
		}
		else if (str[i]>='0'&&str[i]<='9'){
			num*=10.;
			num+=(float)(str[i]-'0');
		}
		else{
			if(str[i]=='E'||str[i]=='e'){
				num=0;
				negFlag=0;
			}
			break;
		}
		i++;
	}
	return num*negFlag;
}

float getFloat(const String& str, int& i){
	str.stb(i);
	float value=atof2(str.c_str()+i);
	str.stwo(i);
	return value;
}

Vec3f Vec3f::normalize()const{
	Vec3f v=*this;
	float dist=sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
	if(dist==0)	return Vec3f(0,0,0);
	float sc=1./dist;

	return v*sc;
}

String Vec3f::toString(){
	String str;
	str+=x;
	str+=' ';
	str+=y;
	str+=' ';
	str+=z;
	return str;
}
Vec3f Rotation::getAxis()const{
	return Vec3f(x,y,z);
}
void rotateVectorX(Vec3f& v,float angle){
	float newY;
	newY=(float)(v.y*cos(angle)-v.z*sin(angle));
	v.z=(float)(v.y*sin(angle)+v.z*cos(angle));
	v.y=newY;
}
void rotateVectorY(Vec3f& v,float angle){
	float newZ;
	newZ=(float)(v.z*cos(angle)-v.x*sin(angle));
	v.x=(float)(v.z*sin(angle)+v.x*cos(angle));
	v.z=newZ;
}
void rotateVectorZ(Vec3f& v,float angle){
	float newX;
	newX=(float)(v.x*cos(angle)-v.y*sin(angle));
	v.y=(float)(v.x*sin(angle)+v.y*cos(angle));
	v.x=newX;
}

Quaternion invert(const Quaternion &srcQuat){
    Quaternion  destQuat;
    float srcQuatNorm;

	srcQuatNorm = 1.0 / (srcQuat.x*srcQuat.x + srcQuat.y*srcQuat.y + srcQuat.z*srcQuat.z + srcQuat.w*srcQuat.w);

	destQuat.x = -srcQuat.x * srcQuatNorm;
	destQuat.y = -srcQuat.y * srcQuatNorm;
	destQuat.z = -srcQuat.z * srcQuatNorm;
	destQuat.w =  srcQuat.w * srcQuatNorm;
	return(destQuat);

}

void rotationToQuaternion(const Rotation &rot, Quaternion &result){
	result.x = rot.x*sin(rot.angle / 2.0);
	result.y = rot.y*sin(rot.angle / 2.0);
	result.z = rot.z*sin(rot.angle / 2.0);
    result.w = cos(rot.angle / 2.0);
}
void quaternionToRotation(Quaternion& q, Rotation& rot){
    rot.angle = acos(q[3]) * 2.0;
	if(rot.angle==0){
		rot=Rotation();
	}
	else{
		rot.x = q[0] / sin(rot.angle / 2.0);
		rot.y = q[1] / sin(rot.angle / 2.0);
		rot.z = q[2] / sin(rot.angle / 2.0);
	}
}
void multiplyQuaternion(Quaternion& q1, Quaternion& q2, Quaternion& result){
    float tmp[4];
    tmp[0] = q2[3] * q1[0] + q2[0] * q1[3] + q2[1] * q1[2] - q2[2] * q1[1];
    tmp[1] = q2[3] * q1[1] + q2[1] * q1[3] + q2[2] * q1[0] - q2[0] * q1[2];
    tmp[2] = q2[3] * q1[2] + q2[2] * q1[3] + q2[0] * q1[1] - q2[1] * q1[0];
    tmp[3] = q2[3] * q1[3] - q2[0] * q1[0] - q2[1] * q1[1] - q2[2] * q1[2];
    result[0] = tmp[0];
	result[1] = tmp[1];
    result[2] = tmp[2];
	result[3] = tmp[3];
}
Rotation Rotation::operator *(const Rotation &par1)const{
	Quaternion q1;
	Quaternion q2;
	Quaternion q3;
	Rotation result;

	rotationToQuaternion(*this, q1);
	rotationToQuaternion(par1, q2);
	multiplyQuaternion(q1, q2, q3);
	quaternionToRotation(q3, result);
	result.setAxis(result.getAxis().normalize());

	return result;
}
Rotation Rotation::operator *=(const Rotation &par1){
	Quaternion q1;
	Quaternion q2;
	Quaternion q3;

	rotationToQuaternion(*this, q1);
	rotationToQuaternion(par1, q2);
	multiplyQuaternion(q1, q2, q3);
	quaternionToRotation(q3, *this);

	return *this;
}

String Rotation::toString()const{
	String str;
	str+=x;
	str+=' ';
	str+=y;
	str+=' ';
	str+=z;
	str+=' ';
	str+=angle;
	return str;
}

Vec3f Rotation::operator *(const Vec3f& vec)const{
		Vec3f axis=getAxis(); //put the axis rotation into the vector axis
	
		//unwind the axis to 1 0 0
		float zAngle=atan2(axis.y,axis.x);	  //get the z angle necessary for unrotating the axis
		rotateVectorZ(axis,-zAngle);

		float yAngle=-atan2( axis.z, axis.x);	//get the y angle necessary for unrotating the axis
		rotateVectorY( axis, -yAngle);
		//axis should allways = 1 0 0 here because we have unwound the vector so that it points to the right, that is the axis the x axis or (1 0 0) 

		//rotate the second VisibleObject only on the x y and x rotations to simulate rotating an object on any given axis to any given angle
		Vec3f retvec=vec;
		rotateVectorZ(retvec,-zAngle);//unwind vec to the x axis
		rotateVectorY(retvec,-yAngle);

		rotateVectorX(retvec, angle);//rotate the vector on the x axis with the angle component exactly the same as if it was rotated on any axis
		
		rotateVectorY(retvec, yAngle);// wind vec back to the specified axis
		rotateVectorZ(retvec, zAngle);
		return retvec;
}
void Rotation::getField(const String& str, int& i){
	Vec3f axis;
	axis.x=getFloat(str, i);
	str.stb(i);
	axis.y=getFloat(str, i);
	str.stb(i);
	axis.z=getFloat(str, i);
	str.stb(i);
	*this=Rotation(axis.normalize(),getFloat(str, i));
}
void Vec3f::getField(const String& str, int& i){
	x=getFloat(str, i);
	str.stb(i);
	y=getFloat(str, i);
	str.stb(i);
	z=getFloat(str, i);
}


String floatToString(float par1){
	char str[0x30];
	sprintf(str,"%.6f",par1);
	removeZeros(str);
	return String(str);
}
void removeZeros(char * str){
	int k;
	if(str[0]=='0'&&str[1]=='.'){
		for(k=0;str[k]!=0;k++){
			str[k]=str[k+1];
		}
		k-=2;
		while(k>0&&str[k]=='0'){
			str[k--]=0;
			if(str[k]=='.'){
				str[k]='0';
				break;
			}
		}
	}
	else if(str[0]=='-'&&str[1]=='0'&&str[2]=='.'){
		for(k=1;str[k]!=0;k++){
			str[k]=str[k+1];
		}
		k-=2;
		while(k>0&&str[k]=='0'){
			str[k--]=0;
			if(str[k]=='.'){
				str[0]='0';
				str[1]=0;
				break;
			}
		}
	}
	else{
		int k=strlen(str)-1;
		while(k>0&&str[k]=='0'){
			str[k--]=0;
			if(str[k]=='.'){
				str[k]=0;
				break;
			}
		}
	}
}
int String::indexOf( String searchString, int position)const{
	const char* match=searchString.pointer->rawChar;
	const char* from=&pointer->rawChar[position];
	int newLength=pointer->rawLen-position;
	for(int i=0;i<newLength;i++){
		if(*from++==*match){
		const char* match2=match+1;
		if(*match2==0)return position+i;
		const char* from2=from;
			for(int k=i;k<newLength;k++){
				if (*match2++!=*from2++)break;
				if(*match2==0)return position+i;
			}
			
		}
	}
	return -1;
}
String& String::operator= (const char& par){
	unsigned lEn=1;

	if(pointer->refCount>1||lEn>pointer->bufLen){
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=createData(lEn+defaultBufLen);
	}
	pointer->rawLen=lEn;
	pointer->rawChar[0]=par;
	return *this;
}
String& String::operator= (float par1){
	char tmp[50];
	sprintf(tmp,"%.6f",par1);
	removeZeros(tmp);
	unsigned lEn=strlen(tmp);

	if(pointer->refCount>1||lEn>pointer->bufLen){
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=createData(lEn+defaultBufLen);
	}
	pointer->rawLen=lEn;
	for(unsigned i=0;i<pointer->rawLen+1;i++)pointer->rawChar[i]=tmp[i];
	return *this;
}
String& String::operator= (int par1){
	char tmp[50];
	sprintf(tmp,"%i",par1);
	unsigned lEn=strlen(tmp);

	if(pointer->refCount>1||lEn>pointer->bufLen){
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=createData(lEn+defaultBufLen);
	}
	pointer->rawLen=lEn;
	for(unsigned i=0;i<pointer->rawLen+1;i++)pointer->rawChar[i]=tmp[i];
	return *this;
}
String &String::operator +=(const String &par1){
	unsigned i;
	unsigned newLength=pointer->rawLen+par1.pointer->rawLen;
	if(pointer->refCount>1||newLength>pointer->bufLen){

		Data* newPointer=createData( newLength+defaultBufLen);
		newPointer->rawLen=pointer->rawLen;
		for(i=0;i<pointer->rawLen+1;i++){
			newPointer->rawChar[i]=pointer->rawChar[i];
		}
		
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=newPointer;
	}

	for(i=0;i<par1.pointer->rawLen+1;i++){
		pointer->rawChar[pointer->rawLen+i]=par1.pointer->rawChar[i];
	}
	pointer->rawLen=newLength;
	return *this;
}
String String::operator +=(char par1){
	Data* newPointer;
	unsigned i;
	if(pointer->refCount>1){
		--pointer->refCount;
		newPointer=createData( pointer->rawLen+defaultBufLen);
		newPointer->rawLen=pointer->rawLen;
		for(i=0; i<pointer->rawLen+1; i++){
			newPointer->rawChar[i]=pointer->rawChar[i];
		}
		pointer=newPointer;
	}
	else if(pointer->rawLen==pointer->bufLen){
		newPointer=createData( pointer->rawLen+defaultBufLen);
		newPointer->rawLen=pointer->rawLen;
		for(i=0; i<pointer->rawLen+1; i++){
			newPointer->rawChar[i]=pointer->rawChar[i];
		}
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=newPointer;
	}
	pointer->rawChar[pointer->rawLen++]=par1;
	pointer->rawChar[pointer->rawLen]=0;
	return *this;
}

inline bool isLegalCppNameChar(const char& ch){
	if( ch>='a'&&ch<='z' ||
		ch>='A'&&ch<='Z' ||
		ch>='0'&&ch<='9' ||
		ch=='_'){
		return true;
	}
	return false;
}

bool String::cmpCppWord(const char* str,int& index)const{
	const char* p2=c_str()+index;
	while(*str!='\0'){
		if(*str++!=*p2++){
			return false;
		};
	}
	if(isLegalCppNameChar(*p2)){
		return false;
	}
	index= p2-c_str();
	return true;
}
bool String::cmpCppWord(const String& str,int& index)const{
	return cmpCppWord(str.pointer->rawChar, index);
}
void String::stb(int& position)const{
	unsigned const char* p=(const unsigned char*)&pointer->rawChar[position];
	int newLength=pointer->rawLen-position-1;
	for(int i=0;i<newLength;i++){
		if(*p>' '&&*p!=','){
			if(*p=='#'){
				while(*p>=' '||*p=='\t')p++;//search for new line
			}
			else{
				break;
			}
		}
		p++;
	}
	position=p-(const unsigned char*)pointer->rawChar;
}
void String::stwo(int& position)const{
	const char* rawChar=pointer->rawChar;
	const char* p=rawChar+position;
	const char* beginP=p;
	const char* endAddress=rawChar+pointer->rawLen;
	while(
		p  <	endAddress	&&
		*p == '_'|| 
		*p == '-' || *p	== '+'|| *p	== '.'|| 	//+ - . needed for legacy reasons ???????
		*p >= '0' && *p	<= '9'||
		*p >= 'A' && *p	<= 'Z'||
		*p >= 'a' && *p	<= 'z'
		){
		p++;
	}
	position=p-rawChar;
}
void String::stw(int& position)const{
	const char* p=&pointer->rawChar[position];
	int newLength=pointer->rawLen-position;
	for(int i=0;i<newLength;i++){
		if(*p<=' '||*p==','){
			position=p-pointer->rawChar;
			return; 
		}
		p++;
	}
	throw this;
}
void Vec3f::toString(String& str){
	str+=x;
	str+=' ';
	str+=y;
	str+=' ';
	str+=z;
}
int hexStrToInt(const String& str, int& i){
	int value=0;
	while(true) {
		char ch=str[i];
		if(ch>='0'&&ch<='9'){
			value<<=4;
			value|=str[i]-'0';
		}
		else if(ch>='a'&&ch<='f'){
			value<<=4;
			value|=str[i]-'a'+10;
		}
		else if(ch>='A'&&ch<='F'){
			value<<=4;
			value|=str[i]-'A'+10;
		}
		else {
			break;
		}
		i++;
	}
	return value;
}
int getInt(const String& str, int& i){
	str.stb(i);
	if(str[i]=='0'){
		if(++i<str.length){
			if(str[i]=='x'||str[i]=='X'){
				return hexStrToInt(str, ++i);
			}
		}
		i--;
	}
	int value=atoi(str.c_str()+i);
	str.stwo(i);
	return value;
}
